/*
  Copyright (C) 2021 Alessandro Languasco & Luca Righi
  (merge of Righi-version 05/03/2021 + computations of new quantities)
	June 2022: max and min |b(chi)| computation inserted
*/

#include "type.h"
#include "computeq.h"
#include <quadmath.h>

#define REAL 0
#define IMAG 1

__float128 Eulergamma = 0.57721566490153286060651209008240243104;
__float128 pi = 3.1415926535897932384626433832795028841971693993751;

/* for odd/even results */
struct result {
  __float128 res;
  __float128 max;
  __float128 min;
  __float128 Kummer;
  // (AL) inserted on 03/02/2021 for computing the class number e Pi(q,1)
  __float128 H1class;
  __float128 Prodresult;
  
  // (AL) 28/05/2021 for computing gamma_Kr, r divisor of (q-1)/2
  __float128 *Dirichletclassresult;
  
  __float128 maxL;
  __float128 minL;
  __float128 maxLprime;
  __float128 minLprime;

  // (AL) inserted on June 6th for computing max and min |b(chi)|
  __float128 maxbchi;
  __float128 minbchi;
  
   // (AL) inserted on Nov 19th 2022 for computing L(1,chi_quad)
  __float128 quadcharvalue;
  
  
};

struct measure {
  __float128 norm2;
  __float128 bnorm2;
  __float128 bnorm2even;
  __float128 bnorm2odd;
  __float128 norm1;
  __float128 bnorm1;
  __float128 bnorm1even;
  __float128 bnorm1odd;
  __float128 norm00;
  __float128 bnorm00;
  __float128 bnorm00even;
  __float128 bnorm00odd;
  __float128 prec2;
  __float128 prec1;
  __float128 prec00;
};

struct norms {
  struct measure norm_ak;
  struct measure norm_lgodd;
  struct measure norm_lgeven;
  struct measure norm_s;
};

/* mmap.c */
void * mapRO_File2Array(const char *filename, int* qd, size_t arraySize, size_t typeSize);
void * mapRWCreate_File2Array(const char* filename, int* qd, size_t arraySize, size_t typeSize);
void * mapRW_File2Array(const char* filename, int* qd, size_t arraySize, size_t typeSize);
void unmap_File2Array(void * mmappedData, size_t filesize, int qd);
char * mallocFilename(const char * basename, const char* suffix);
void * mkmmapcopy(void * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix, const char mode, int *qd);
fftwq_complex * mapfile(const char * name, const char * basename, long int size, int * qd);
void  print_time(const char * message, double startplan, double endplan);
void * coniugateAndDivideFFTFileInMappedFile(fftwq_complex * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix);
void * copyAndNormalizeFFTFileInMappedFile(fftwq_complex * a, ptrdiff_t n, size_t typesize, const char * basename, const char * suffix);
void printcomplexvector(fftwq_complex * v, ptrdiff_t start, ptrdiff_t step, ptrdiff_t size, const char * name, FILE *stream);
/* s.c */
long int lpow_mod(long int base, long int exp, long int mod);
__float128 s(__float128 aq, int precision, int verbose);

struct result readResult(const char * basename, const char * name) {
  const char * funcname = "readResult()";
  char buf[1024];
  
  struct result * r = NULL;
  if ((r = malloc(sizeof(struct result))) == NULL) {
    fprintf(stderr, "%s: unable to allocate result. Giving up\n", funcname);
    exit(1);
  }
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "%s: can't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->res)) != 1)) {
    fprintf(stderr, "%s: unable to read 'res' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->max)) != 1)) {
    fprintf(stderr, "%s: unable to read 'max' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->min)) != 1)) {
    fprintf(stderr, "%s: unable to read 'min' from %s: %s\n", funcname, filename, strerror(errno));
  }

  /* */
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->Kummer)) != 1)) {
    fprintf(stderr, "%s: unable to read 'Kummer' from %s: %s\n", funcname, filename, strerror(errno));
  }

  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->maxL)) != 1)) {
    fprintf(stderr, "%s: unable to read 'maxL' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->minL)) != 1)) {
    fprintf(stderr, "%s: unable to read 'minL' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->maxLprime)) != 1)){
    fprintf(stderr, "%s: unable to read 'maxLprime' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->minLprime)) != 1)){
    fprintf(stderr, "%s: unable to read 'minLprime' from %s: %s\n", funcname, filename, strerror(errno));
  }

  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->maxbchi)) != 1)){
    fprintf(stderr, "%s: unable to read 'maxbchi' from %s: %s\n", funcname, filename, strerror(errno));
  }
  
  if ((fscanf(f, "%s\n", buf) != 1) || (readf(buf, &(r->minbchi)) != 1)){
    fprintf(stderr, "%s: unable to read 'minbchi' from %s: %s\n", funcname, filename, strerror(errno));
  }  
  /* */   

  fclose(f);

  // printf("%s: readed %Lf, %Lf, %Lf from %s\n", funcname, r->res, r->max, r->min, filename);
  printf("%s: read %s, %s, %s, %s, %s, %s, %s, %s, %s, %s from %s\n", funcname, 
	 fstr(r->res), fstr(r->max), fstr(r->min), fstr(r->Kummer), fstr(r->maxL), fstr(r->minL),
	 fstr(r->maxLprime), fstr(r->minLprime), fstr(r->maxbchi), fstr(r->minbchi), filename);
  
  return *r;
}

void writeResult(struct result r, const char * basename, const char * name) {

  const char * funcname = "writeResult()";
  
  char * filename = mallocFilename(basename, name);
  FILE * f = NULL;
  if ((f = fopen(filename, "w")) == NULL) {
    fprintf(stderr, "%s: san't open %s: %s\n", funcname, filename, strerror(errno));
    exit(1);
  }

  fprintf(f, "%s\n", fstr(r.res));
  fprintf(f, "%s\n", fstr(r.max));
  fprintf(f, "%s\n", fstr(r.min)); 
  fprintf(f, "%s\n", fstr(r.Kummer));
  fprintf(f, "%s\n", fstr(r.maxL));
  fprintf(f, "%s\n", fstr(r.minL));
  fprintf(f, "%s\n", fstr(r.maxLprime));
  fprintf(f, "%s\n", fstr(r.minLprime));
  fprintf(f, "%s\n", fstr(r.maxbchi));
  fprintf(f, "%s\n", fstr(r.minbchi));

  fclose(f);

  return;
}

fftwq_complex * mkfftin(ptrdiff_t nhalf, const char *basename, int *fd_fftin) {

  char * funcname = "mkfftin()";
  char * fftinFilename = mallocFilename(basename, N_FFTINPUT);

  fftwq_complex * fftin = (fftwq_complex *) mapRWCreate_File2Array(fftinFilename, fd_fftin, nhalf, sizeof(fftwq_complex));
  
  if (fftin == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinFilename);
    return NULL;
  }

  return fftin;
}

fftwq_complex * mkfftout(ptrdiff_t nhalf, const char *basename, int *fd_fftout) {

  char * funcname = "mkfftout()";
  char * fftoutFilename = mallocFilename(basename, N_FFTOUTPUT);

  fftwq_complex * fftout = (fftwq_complex *) mapRWCreate_File2Array(fftoutFilename, fd_fftout, nhalf, sizeof(fftwq_complex));
  if (fftout == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftoutFilename);
    return 0;
  }

  return fftout;
}

fftwq_complex * mkfftinverse(ptrdiff_t nhalf, const char *basename, int *fd_fftinverse) {

  char * funcname = "mkfftinverse()";
  char * fftinverseFilename = mallocFilename(basename, N_FFTINVERSE);

  fftwq_complex * fftinverse = (fftwq_complex *) mapRWCreate_File2Array(fftinverseFilename, fd_fftinverse, nhalf, sizeof(fftwq_complex));
  if (fftinverse == NULL) {
    fprintf(stderr, "%s: unable to map file(%s)\n", funcname, fftinverseFilename);
    return 0;
  }

  return fftinverse;
}


fftwq_plan mkplan(fftwq_complex *fftin, fftwq_complex *fftout, ptrdiff_t nhalf) {

  if (fftin == fftout) {
    /* same array. Wrong. This is an outplace FFT */
    fprintf(stderr,  "*************************************************\n");
    fprintf(stderr,  "*** ERROR: fftin == fftout. FFT in not INPLACE. *\n");
    fprintf(stderr,  "*************************************************\n");
    return NULL;
  }
  
  clock_t startplan = clock();
  
  if (fftwq_import_wisdom_from_filename("wisdom-outplace-fftwq-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-outplace-fftwq-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-outplace-fftwq-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwq_plan plan = fftwq_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for OUTPLACE transforms*/ 
              fftout,  /*output for OUTPLACE transforms*/ 
              FFTW_FORWARD, /* -1 because it is the sign of FORWARD transform */
              FFTW_ESTIMATE); /* flags */
              
  if (plan == NULL) { 
    fprintf(stderr,"plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwq_import_wisdom_from_filename("wisdom-outplace-fftwq-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-outplace-fftwq-estimate.txt \n");
    fftwq_export_wisdom_to_filename("wisdom-outplace-fftwq-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Plan generation", startplan, endplan);

  return plan;
}

fftwq_plan mkinverseplan(fftwq_complex *fftin, fftwq_complex *fftout, ptrdiff_t nhalf) {

  if (fftin == fftout) {
    /* same array. Wrong. This is an outplace FFT */
    fprintf(stderr,  "*************************************************\n");
    fprintf(stderr,  "*** ERROR: fftin == fftout. FFT in not INPLACE. *\n");
    fprintf(stderr,  "*************************************************\n");
    return NULL;
  }
  
  clock_t startplan = clock();
  
  if (fftwq_import_wisdom_from_filename("wisdom-outplace-inverse-fftwq-estimate.txt")) {
    fprintf(stderr,  "*** Reading wisdom-outplace-inverse-fftwq-estimate \n");
  } else {
    fprintf(stderr,  "*** Generating wisdom-outplace-inverse-fftwq-estimate \n");
  }

  int rank=1;
  fftw_iodim64 * dims = malloc(rank * sizeof(fftw_iodim64));
  dims[0].n= nhalf;   /* length of the trasform */
  dims[0].is=1;
  dims[0].os=1; 
  int howmany_rank=1; 
  fftw_iodim64 * howmany_dims = malloc(howmany_rank * sizeof(fftw_iodim64)); 
  howmany_dims[0].n=1;
  howmany_dims[0].is=1;
  howmany_dims[0].os=1;
  
  fftwq_plan plan = fftwq_plan_guru64_dft(rank, /*rank=1 means monodimensional vector*/
              dims, /* nhalf */
              howmany_rank,
              howmany_dims,
              fftin,   /*input for OUTPLACE transforms*/ 
              fftout,  /*output for OUTPLACE transforms*/ 
              FFTW_BACKWARD, /* 1 because it is the sign of BACKWARD transform */
              FFTW_ESTIMATE); /* flags */
              
  if (plan == NULL) { 
    fprintf(stderr,"inverse plan creation failed\n");
    return 0;
  }
    
  /*  saving  a  generated  wisdom  for  future  runs  */
  if (!fftwq_import_wisdom_from_filename("wisdom-outplace-inverse-fftwq-estimate.txt")) {
    fprintf(stderr,  "***  Saving wisdom-outplace-inverse-fftwq-estimate.txt \n");
    fftwq_export_wisdom_to_filename("wisdom-outplace-inverse-fftwq-estimate.txt"); 
  }
  
  clock_t endplan = clock(); 

  print_time("Inverse plan generation time", startplan, endplan);

  return plan;
}

fftwq_complex * mkLGodd(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_LGodd) {

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 sinvalue;
  __float128 cosvalue;
  __float128 logpi = logq(pi);
  __float128 twist = 0.0Q;
  __float128 angle = 0.0Q; 
  __float128 aux;
  __float128 d;
  long int a;
  ptrdiff_t k;
  
  /* 
   * STEP: mkLGodd
   *  need: fftin mmapped array
   *  provide: file <basename>-LGodd.qd
   */
  
  clock_t startclock = clock();  /* start clock */

  angle = (__float128) pi / nhalf;
  twist = 0.0Q; 
  a = 1; 

  for (k = 0; k <nhalf ; k++){   
     cosvalue = cosq(twist);
     sinvalue = sinq(twist);  
     d = (__float128) a/q ;
     aux = (2.0Q) * lgammaq( d ) + logq( sinq(pi * d)) - logpi ;
     fftin[k][REAL] = aux * cosvalue;  
     fftin[k][IMAG] = (-1.0Q) * aux * sinvalue; 
     a = (a * primroot) % q;  
     twist = twist + angle;
  }

  clock_t endclock = clock();  /* end clock */  
  print_time("LNGammaOdd computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * lgodd = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_LGODD, 'r', fd_LGodd);

  return lgodd;

}

fftwq_complex * fftLGodd(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_LGoddfft) {
  
  /* 
   * STEP: fftLGodd
   *  need: fftin mmapped array with LGodd data
   *  provide: fft of fftin in array fftout
   *  provide: copy of fftout in file <basename>-LGoddfft.qd if basename != NULL
   */

  clock_t startfft = clock(); /* start clock */
  fftwq_execute(plan); 
  clock_t endfft = clock(); /* stop clock */
  print_time("FFT(LGodd)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * lgoddfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTLGODD, 'r', fd_LGoddfft);

  return lgoddfft;

}

/* to be developed */
fftwq_complex * fft_INV(fftwq_complex *fftin, fftwq_complex *fftout,
			fftwq_plan iplan, long int nhalf,
			const char *basename,
			const char *fftname,
			const char *fftinvname,
			int *fd_INV) {
  
  /* 
   * STEP: fft_INV
   *  need: fftin mmapped array to be filled with fft data (from fftname)
   *  need: basename: basename for the files
   *  need: iplan: plan for the inverse FFT
   *  need: fftname: suffix for basename to open th file with the input fft data
   *  need: fftinvname: suffix for basename to write the file with the fft inverse data
   *  provide: fft(fftin) in array fftout 
  *  provide: copy of fftout in file <basename>-<fftinvname>
   */

  char* funcName = "FFT_INV()";

  clock_t start_bcopy = clock(); /* start clock */
  if ((fftin = copyAndNormalizeFFTFileInMappedFile(fftin, nhalf, sizeof(fftwq_complex), basename, fftname)) == NULL) {
    fprintf(stderr, "%s: *** ERROR: unable to loadFilenameInMappedFile(%s,%s)\n", funcName, basename, fftname);
    exit(0);    
  }
  clock_t end_bcopy = clock(); /* stop clock */
  print_time("FFT_INV(): Initial copy and normalization time\n", start_bcopy, end_bcopy);
  
  clock_t startfft_INV = clock(); /* start clock */
  fftwq_execute(iplan); 
  clock_t endfft_INV = clock(); /* stop clock */
  print_time("FFT_INV(): fftwq inverse plan execute time", startfft_INV, endfft_INV);
  fflush(stdout);

  clock_t start_copy = clock(); /* start clock */
  /* make the copy of the output in basename-fftinvname */
  mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, fftinvname, '-', fd_INV);
  clock_t end_copy = clock(); /* stop clock */
  print_time("FFT_INV(): Result copy", start_copy, end_copy);
  
  return fftout;

}

/* FOR CHECKING COMPUTATION ACCURACY; see LR - MATH COMP - SECT 5.7
 * compute Norm00, Norm1 and Norm2 for |base - inverse| / |base|
 */
int computenorms(const char *basename, const char *base, const char *inverse, long int size, struct measure *m, int evenodd) {
  const char* funcname = "norm2diff()";
  fftwq_complex * b = NULL;
  fftwq_complex * i = NULL;
  __float128 x, y;
  int fd_b, fd_i;

  if ((b = mapfile(base, basename, size, &fd_b)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, base);
    return 0;
  }

  if ((i = mapfile(inverse, basename, size, &fd_i)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, inverse);
    return 0;
  }

  /* NORM2 sum vars for ||b - i|| */
  __float128 SUM2 = 0.0Q;

  /* NORM2 sum vars for ||b|| (the original vector) */
  __float128 bSUM2 = 0.0Q;

  /* NORM1 sum vars for ||b - i|| */
  __float128 SUM1 = 0.0Q;
  
  /* NORM1 sum vars for ||b|| (the original vector) */
  __float128 bSUM1 = 0.0Q;
  
  /* NORM00 vars for ||b - i|| */
  __float128 DMAX = 0.0Q;
  __float128 DTEMP = 0.0Q;

  /* NORM00 vars for ||b|| */
  __float128 BMAX = 0.0Q;
  __float128 BTEMP = 0.0Q;

  for (long int k = 0; k < size; k++) {
    /* difference vector component */
    x = b[k][REAL] - i[k][REAL];
    y = b[k][IMAG] - i[k][IMAG];

    printf("%s + %s%s ", fstr(x), fstr(y), "i");

    /* NORM2 b-i */
    SUM2 += (x * x + y * y);

    /* NORM2 b */
    bSUM2 += (b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);

    /* NORM1 b-i */
    SUM1 += sqrtq(x * x + y * y);

    /* NORM1 b */
    bSUM1 += sqrtq(b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);

    /* NORM00 b - i */
    DTEMP = sqrtq(x * x + y * y);
    if (DTEMP > DMAX) {
      DMAX = DTEMP;
      // DPOS = k;
    }

    /* NORM00 b */
    BTEMP = sqrtq(b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);
    if (BTEMP > BMAX) {
      BMAX = BTEMP;
      // BPOS = k;
    }
  }

  printf("]\n");
  
  m->norm2 = sqrtq(SUM2);
  m->bnorm2 = sqrtq(bSUM2);

  m->norm1 = SUM1;
  m->bnorm1 = bSUM1;

  // printf("DEBUG: norm2 = %32.30LE, norm1 = %32.30LE\n", m->norm2, m->norm1);
  
  m->norm00 = DMAX;
  m->bnorm00 = BMAX;

  if (fabsq(m->bnorm2) > (0.000001Q)) {
    m->prec2 = m->norm2 / m->bnorm2;
  } else {
    m->prec2 = 0.0Q;
  }

  if (fabsq(m->bnorm1) > (0.000001Q)) {
    m->prec1 = m->norm1 / m->bnorm1;
  } else {
    m->prec1 = 0.0Q;
  }

  if (fabsq(m->bnorm00) > (0.000001Q)) {
    m->prec00 = m->norm00 / m->bnorm00;
  } else {
    m->prec00 = 0.0Q;
  }
  
  unmap_File2Array(b, size, fd_b);
  unmap_File2Array(i, size, fd_i);

  return 1;
}

void nullmeasure(struct measure *m) {
  if (m == NULL) {
    fprintf(stderr, "nullmeasure(): measure == NULL\n");
    return;
  }

  m->norm2 = 0.0Q;
  m->bnorm2 = 0.0Q;
  m->bnorm2even = 0.0Q;
  m->bnorm2odd = 0.0Q;
  m->norm1 = 0.0Q;
  m->bnorm1 = 0.0Q;
  m->bnorm1even = 0.0Q;
  m->bnorm1odd = 0.0Q;
  m->norm00 = 0.0Q;
  m->bnorm00 = 0.0Q;
  m->bnorm00even = 0.0Q;
  m->bnorm00odd = 0.0Q;
  m->prec2 = 0.0Q;
  m->prec1 = 0.0Q;
  m->prec00 = 0.0Q;

}

/* FOR CHECKING COMPUTATION ACCURACY; see LR - MATH COMP - SECT 5.7
 * compute Norm00, Norm1 and Norm2 for |base - inverse| / |base|
 */
int computenormskahan(const char *basename, const char *base, const char *inverse, long int size, struct measure *m, int evenodd) {
  const char* funcname = "norm2diff()";
  fftwq_complex * b = NULL;
  fftwq_complex * i = NULL;
  __float128 x, y, belem;
  int fd_b, fd_i;

  int DEBUG=0;

  if ((b = mapfile(base, basename, size, &fd_b)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, base);
    nullmeasure(m);
    return 0;
  }

  if ((i = mapfile(inverse, basename, size, &fd_i)) == NULL) {
    fprintf(stderr, "%s: unable to map file %s%s\n", funcname, basename, inverse);
    nullmeasure(m);
    return 0;
  }

  /* NORM2 Kahan sum vars for ||b - i|| */
  __float128 SUM2 = 0.0Q;
  __float128 C2 = 0.0Q;
  __float128 Y2 = 0.0Q;
  __float128 T2 = 0.0Q;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) */
  __float128 bSUM2 = 0.0Q;
  __float128 bC2 = 0.0Q;
  __float128 bY2 = 0.0Q;
  __float128 bT2 = 0.0Q;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) EVEN */
  __float128 bSUM2e = 0.0Q;
  __float128 bC2e = 0.0Q;
  __float128 bY2e = 0.0Q;
  __float128 bT2e = 0.0Q;

  /* NORM2 Kahan sum vars for ||b|| (the original vector) ODD */
  __float128 bSUM2o = 0.0Q;
  __float128 bC2o = 0.0Q;
  __float128 bY2o = 0.0Q;
  __float128 bT2o = 0.0Q;

  /* NORM1 Kahan sum vars for ||b - i|| */
  __float128 SUM1 = 0.0Q;
  __float128 C1 = 0.0Q;
  __float128 Y1 = 0.0Q;
  __float128 T1 = 0.0Q;
  
  /* NORM1 Kahan sum vars for ||b|| (the original vector) */
  __float128 bSUM1 = 0.0Q;
  __float128 bC1 = 0.0Q;
  __float128 bY1 = 0.0Q;
  __float128 bT1 = 0.0Q;

  /* NORM1 Kahan sum vars for ||b|| (the original vector) EVEN */
  __float128 bSUM1e = 0.0Q;
  __float128 bC1e = 0.0Q;
  __float128 bY1e = 0.0Q;
  __float128 bT1e = 0.0Q;

  /* NORM1 Kahan sum vars for ||b|| (the original vector) ODD */
  __float128 bSUM1o = 0.0Q;
  __float128 bC1o = 0.0Q;
  __float128 bY1o = 0.0Q;
  __float128 bT1o = 0.0Q;
  
  /* NORM00 vars for ||b - i|| */
  __float128 DMAX = 0.0Q;
  __float128 DTEMP = 0.0Q;

  /* NORM00 vars for ||b|| */
  __float128 BMAX = 0.0Q;
  __float128 BTEMP = 0.0Q;

  /* NORM00 vars for ||b|| EVEN */
  __float128 BMAXe = 0.0Q;

  /* NORM00 vars for ||b|| ODD */
  __float128 BMAXo = 0.0Q;

  if (DEBUG) {
    printcomplexvector(b, 0, 1, size, base, stdout); // complete vector
    printcomplexvector(b, 0, 2, size, base, stdout); // even components
    printcomplexvector(b, 1, 2, size, base, stdout); // odd component
  }
  
  for (long int k = 0; k < size; k++) {

    /* difference vector component */
    x = b[k][REAL] - i[k][REAL];
    y = b[k][IMAG] - i[k][IMAG];

    /* NORM2 b-i */
    Y2 = (x * x + y * y) - C2;   /* qui sommo i quadrati dei valori assoluti delle componenti poi dovro` fare la radice quadrata */
    T2 = SUM2 + Y2;
    C2 = (T2 - SUM2) - Y2;
    SUM2 = T2;

    /* NORM2 b */
    belem = (b[k][REAL] * b[k][REAL] + b[k][IMAG] * b[k][IMAG]);
    bY2 = belem - bC2;
    bT2 = bSUM2 + bY2;
    bC2 = (bT2 - bSUM2) - bY2;
    bSUM2 = bT2;

    if (evenodd && (k % 2 == 0)) {
      /* NORM2 b EVEN */
      bY2e = belem - bC2e;
      bT2e = bSUM2e + bY2e;
      bC2e = (bT2e - bSUM2e) - bY2e;
      bSUM2e = bT2e;
    }
    
    if (evenodd && (k % 2 == 1)) {
      /* NORM2 b ODD */
      bY2o = belem - bC2o;
      bT2o = bSUM2o + bY2o;
      bC2o = (bT2o - bSUM2o) - bY2o;
      bSUM2o = bT2o;
    }

    /* NORM1 b-i */
    Y1 = sqrtq(x * x + y * y) - C1; /* qui sommo i valori assoluti delle componenti */
    T1 = SUM1 + Y1;
    C1 = (T1 - SUM1) - Y1;
    SUM1 = T1;

    /* NORM1 b */
    bY1 = sqrtq(belem) - bC1;
    bT1 = bSUM1 + bY1;
    bC1 = (bT1 - bSUM1) - bY1;
    bSUM1 = bT1;

    if (evenodd && (k % 2 == 0)) {
      /* NORM1 b EVEN */
      bY1e = sqrtq(belem) - bC1e;
      bT1e = bSUM1e + bY1e;
      bC1e = (bT1e - bSUM1e) - bY1e;
      bSUM1e = bT1e;
    }
    
    if (evenodd && (k % 2 == 1)) {
      /* NORM1 b ODD */
      bY1o = sqrtq(belem) - bC1o;
      bT1o = bSUM1o + bY1o;
      bC1o = (bT1o - bSUM1o) - bY1o;
      bSUM1o = bT1o;
    }

    /* NORM00 b - i */
    DTEMP = sqrtq(x * x + y * y);
    if (DTEMP > DMAX) {
      DMAX = DTEMP;
    }

    /* NORM00 b */
    BTEMP = sqrtq(belem);
    if (BTEMP > BMAX) {
      BMAX = BTEMP;
    }

    if (evenodd && (k % 2 == 0) && (BTEMP > BMAXe)) {
      BMAXe = BTEMP;
    }

    if (evenodd && (k % 2 == 1) && (BTEMP > BMAXo)) {
      BMAXo = BTEMP;
    }

  }

  m->norm2 = sqrtq(SUM2);
  m->bnorm2 = sqrtq(bSUM2);
  if (evenodd) {
    m->bnorm2even = bSUM2e;
    m->bnorm2odd = bSUM2o;
  }

  m->norm1 = SUM1;
  m->bnorm1 = bSUM1;
  if (evenodd) {
    m->bnorm1even = bSUM1e;
    m->bnorm1odd = bSUM1o;
  }
  
  m->norm00 = DMAX;
  m->bnorm00 = BMAX;
  if (evenodd) {
    m->bnorm00even = BMAXe;
    m->bnorm00odd = BMAXo;
  }

  if (fabsq(m->bnorm2) > 0.000001) {
    m->prec2 = m->norm2 / m->bnorm2;
  } else {
    m->prec2 = 0.0Q;
  }

  if (fabsq(m->bnorm1) > 0.000001) {
    m->prec1 = m->norm1 / m->bnorm1;
  } else {
    m->prec1 = 0.0Q;
  }

  if (fabsq(m->bnorm00) > 0.000001) {
    m->prec00 = m->norm00 / m->bnorm00;
  } else {
    m->prec00 = 0.0Q;
  }
  
  unmap_File2Array(b, size, fd_b);
  unmap_File2Array(i, size, fd_i);

  return 1;
}

fftwq_complex * mkAK(fftwq_complex *fftin, ptrdiff_t nhalf, long int q, long int primroot, const char * basename, int * fd_AK) {

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 sinvalue;
  __float128 cosvalue;
  __float128 twist = 0.0Q;
  __float128 angle = 0.0Q; 
  __float128 c;
  __float128 lq = (__float128) q;
  ptrdiff_t a;
  ptrdiff_t k;

  /* 
   * STEP: mkAK
   *  need: fftin mmapped array available for writing
   *  provide: input data AK (in fftin) to compute <basename>-AKfft.qd
   */
  
  clock_t startclock = clock();  /* start clock */  
  angle = (__float128) pi/nhalf;
  twist = 0.0Q; 
  a = 1; 
  for (k = 0; k < nhalf ; k++){
    /*g^k -R g^(k+m) \equiv g^k  -R (q-1.0Q)g^k \equiv g^k -R (-g^k) \equiv g^k -R (q -g^k) = 2*g^k -R q */

    /* -- 20210107 -- ERA: c = (__float128) (2*a-q); Vedi mail del 20210102 */ 
    // c = (__float128) ((2.0 * ((__float128)a) / ((__float128)q)) - 1.0Q);
    c = (__float128) (2.0Q * (__float128) a - lq) / lq;
    // c = (__float128) a/q;
    // c = (__float128) 2.0Q * c - 1.0Q;
    cosvalue = cosq(twist);
    sinvalue = - sinq(twist);  /* depends on the sign = -1 of the transform */
    fftin[k][REAL] = c * cosvalue;
    fftin[k][IMAG] = c * sinvalue;
    a = (a * primroot) % q;
    twist += angle;
  }
  clock_t endclock = clock();  /* end clock */
  print_time("AK computation", startclock, endclock);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /* copy needed */
  fftwq_complex * ak = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_AK, 'r', fd_AK);

  return ak;

}

fftwq_complex * fftAK(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_AKfft) {
  
  /* STEP: fftAK 
   *  need: fftin mmapped array with ak start data 
   *  provide: fft in fftout, file <basename>-AKoddfft.qd if basename != NULL
   */

  clock_t startfft = clock();
  fftwq_execute(plan);
  clock_t endfft  = clock();
  print_time("FFT(AK)", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  /*
  ptrdiff_t q = 2 * nhalf + 1;
  for (ptrdiff_t k = 0; k < nhalf; k++) {
    fftout[k][REAL] *= q;
    fftout[k][IMAG] *= q;
  }
  */
  
  fftwq_complex * akfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTAK, 'r', fd_AKfft);

  return akfft;

}

// AL inserted 28/05/2021 for the divisors; reads the sorted array saved by the main program
void getDivisors(ptrdiff_t *divisors, ptrdiff_t NUM_DIV) {
  char * funcName = "getDivisors()";
  FILE *myFile = NULL;
  ptrdiff_t i;
  ptrdiff_t dummy;
  int got;
  
  if ((myFile = fopen("divfile.res", "r")) == NULL) {
    fprintf(stderr, "%s: unable to open \"divfile.res\": %s\n", funcName, strerror(errno));
    return;
  }
  
  if ((got = fscanf(myFile, "%tu", &dummy)) < 1) {
    fprintf(stderr, "%s: unable to read first line from \"divfile.res\"\n", funcName);
    return;
  }
  
  for (i = 0; i < NUM_DIV; i++){
    if ((got = fscanf(myFile, "%tu", &divisors[i])) != 1) {
      fprintf(stderr, "%s: unable to read %ld line from \"divfile.res\"\n", funcName, i);
    }
  } 
  
  fclose(myFile);
    
}

// AL inserted 28/05/2021 for the divisors; it is just the binary search on a sorted array

ptrdiff_t binarySearch(ptrdiff_t arr[], ptrdiff_t l, ptrdiff_t r, ptrdiff_t x)
{
ptrdiff_t m = 0;
  while (l <= r)
  {
     m = l + (r-l)/2;
  
    // Check if x is present at mid
    if (arr[m] == x) 
        return m;  
  
    // If x greater, ignore left half  
    if (arr[m] < x) 
        l = m + 1; 
  
    // If x is smaller, ignore right half 
    else 
         r = m - 1; 
  }
  
  // if we reach here, then element was not present
  return -1; 
}

// AL inserted 28/05/2021 for the divisors; just prints an array of integers

void stampaint(ptrdiff_t* result, ptrdiff_t NUM_POINTS) { /* for checking results */
    int i;
    for (i = 0; i < NUM_POINTS; ++i) {
        printf("%ld\n", result[i]);
    }
}

struct result oddResult(fftwq_complex *lngammaodd_fft, fftwq_complex *ak_fft, ptrdiff_t nhalf, ptrdiff_t q, ptrdiff_t NUM_DIV, ptrdiff_t* divisors) {

  struct result myresult; 
  
  char bufRe[512];
  printf("%s", "Computing the sum and minmax over odd characters"); 
  fflush(stdout);
  
  clock_t startcompodd = clock(); /* start clock */

  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  __float128 oddresult = 0;
  __float128 res;
  __float128 lq = (__float128) q;
  __float128 akREq;
  __float128 akIMq;
  __float128 Kummer_res;
  __float128 Kummer;
  // (AL) inserted on 03/02/2021 -- needed to compute class number h1 and Pi(q,1)
  __float128 H1class;
  __float128 Prododdresult;   
  
  __float128 realpart;
  __float128 imagpart;
  __float128 Minimumoddresult;
  __float128 Maximumoddresult;
  __float128 MinimumLoddresult;
  __float128 MaximumLoddresult;
  __float128 MinimumLprimeoddresult;
  __float128 MaximumLprimeoddresult;
  __float128 Minimumbchioddresult;
  __float128 Maximumbchioddresult;  
  __float128 Minimum = FLT128_MAX;/* odd characters */
  __float128 Maximum = 0; 
  __float128 MinimumL = FLT128_MAX;/* odd characters */
  __float128 MaximumL = 0;
  __float128 MinimumLprime = FLT128_MAX;/* odd characters */
  __float128 MaximumLprime = 0; 
  __float128 Minimumbchi = FLT128_MAX;/* odd characters */
  __float128 Maximumbchi = 0.0Q;  
  
  __float128 quadcharvalue = 0.0Q;
    
  long int jmin = 0;
  long int jmax = 0;
  long int jminL = 0;
  long int jmaxL = 0;
  long int jminLprime = 0;
  long int jmaxLprime = 0;
  long int jminbchi = 0;
  long int jmaxbchi = 0;  
  __float128 num     = 0;
  __float128 Lquotient = 0;
  __float128 num1    = 0;
  __float128 den     = 0;
  __float128 quot    = 0;   
  __float128 aux     = 0; 
  __float128 absbchi = 0;
  __float128 bchiREq = 0;
  __float128 bchiIMq = 0;  
  __float128 frac    = 0;   
  __float128 frac1   = 0;   
  __float128 corr    = logq((2.0Q)*pi)+Eulergamma;  
  // __float128 corrq   = corr/q; // corr/q;
  __float128 Kummer_realcorrection = (logq(pi) - (1.5Q) * logq((__float128) q)) * nhalf;
  
  // (AL) inserted on 03/02/2021 -- needed to compute the class number h_1(q)
  __float128 H_correction = - (nhalf -1.0Q) * (logq((__float128) q) + logq(2.0Q));
  
    // (AL) inserted on 06/06/2022 -- for computing b_chi  
  __float128 minuslogq    = -logq(q);  
  

  res = 0.0Q;
  Kummer_res = 0.0Q;
  oddresult = 0.0Q;
  
  // 28/05/2021 for computing the sums over Dirichlet character of index ell divided by r
  __float128 *Oddclasssumr;
  Oddclasssumr = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV);
  for (ptrdiff_t  j = 0; j < NUM_DIV; j++) { Oddclasssumr[j] = 0.0Q ;};
  
  myresult.Dirichletclassresult = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV);
  ptrdiff_t r;
   //

  /* computes |(a+ib)/(c+id)|= |a+ib|/|c+id| and the real part of (a+ib)/(c+id); the imaginary part is 0 by def */
  
  for (long int i = 0;  i < nhalf; ++i) {
    // -- code replaced LR 20210115
    // num = q * (lngammaodd_fft[i][REAL] * ak_fft[i][REAL] + lngammaodd_fft[i][IMAG] * ak_fft[i][IMAG]);
    // den = q * q * (ak_fft[i][REAL] * ak_fft[i][REAL] + ak_fft[i][IMAG] * ak_fft[i][IMAG]);
    akREq = lq * ak_fft[i][REAL];
    akIMq = lq * ak_fft[i][IMAG];
    num =  (lngammaodd_fft[i][REAL] * akREq + lngammaodd_fft[i][IMAG] * akIMq); 
    den =  (akREq * akREq + akIMq * akIMq);
    // -- end replaced code
    
    Lquotient = num / den;
    // 28/05/2021 for gamma_Kr
    Oddclasssumr[0] = Oddclasssumr[0]+ Lquotient;
    res = Oddclasssumr[0];

    for (ptrdiff_t j = 1; j < NUM_DIV; j++) { r = divisors[j];
        					 // r | char index  
           		 if ((2*i+1)%r == 0) {Oddclasssumr[j] = Oddclasssumr[j]+ Lquotient ;}; 
                }; 
                
    realpart = corr * ak_fft[i][REAL] + lngammaodd_fft[i][REAL];
    imagpart = corr * ak_fft[i][IMAG] + lngammaodd_fft[i][IMAG];
    aux = realpart * realpart + imagpart * imagpart;
    Kummer_res = Kummer_res + (0.5Q) * logq(den);
    // -- code replaced: LR 20210115
    // num1 = q * (lngammaodd_fft[i][IMAG] * ak_fft[i][REAL] - lngammaodd_fft[i][REAL] * ak_fft[i][IMAG]);
    num1 = (lngammaodd_fft[i][IMAG] * akREq - lngammaodd_fft[i][REAL] * akIMq);
    // -- end replaced code
    // 06/06/2022 for bchi
    bchiREq = (-q)*num/den + minuslogq;
    bchiIMq = q*num1/den; 
    absbchi = bchiREq * bchiREq + bchiIMq * bchiIMq;

    frac = corr + q * Lquotient;
    frac1 = q * num1 / den;
    quot = frac * frac + frac1 * frac1;
    if (quot < Minimum) { Minimum = quot; jmin = 2*i+1;};
    if (quot > Maximum) { Maximum = quot; jmax = 2*i+1;};
    if (den < MinimumL) { MinimumL = den; jminL = 2*i+1;}; 
    if (den > MaximumL) { MaximumL = den; jmaxL = 2*i+1;};
    if (aux < MinimumLprime) { MinimumLprime = aux; jminLprime = 2*i+1;};
    if (aux > MaximumLprime) { MaximumLprime = aux; jmaxLprime = 2*i+1;};
    if (absbchi < Minimumbchi) { Minimumbchi = absbchi; jminbchi = 2*i+1 ;};
    if (absbchi > Maximumbchi) { Maximumbchi = absbchi; jmaxbchi = 2*i+1 ;};
    
   }
  
   if (nhalf % 2 == 1)
  		{ ptrdiff_t quad_index = floor(nhalf/2);
		  quadcharvalue = fabsq(pi* ak_fft[quad_index][REAL]/sqrtq(lq));
		};
  
  Minimumoddresult = sqrtq(Minimum); /* min L'/L for odd characters */
  Maximumoddresult = sqrtq(Maximum); /* max L'/L for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L'/L (1,chi) \n ");
  printf("chiminodd index = ");
  printf("%ld\n",jmin);
  printf("chimaxodd index = ");
  printf("%ld\n",jmax);
  
  oddresult = q * res; /* correction factors for odd characters */

  fprintf(stdout, "Kummer_realcorrection = %s\n", fstr(Kummer_realcorrection));  
  fprintf(stdout, "Kummer_res = %s\n", fstr(Kummer_res));
  Kummer = expq(Kummer_res + Kummer_realcorrection);
  fprintf(stdout, "Kummer (r) = %s\n", fstr(Kummer));
  
  
  // (AL) inserted on 03/02/2021 -- prints also log(Kummer)
  fprintf(stdout, "log r = %s\n", fstr(Kummer_res + Kummer_realcorrection));
  // (AL) inserted on 03/02/2021 -- needed to compute the class number h_1(q)
  fprintf(stdout, "H_correction = %s\n", fstr(H_correction));   
  H1class = Kummer_res + H_correction;
  fprintf(stdout, "log h1 = %s\n", fstr(H1class));

  // (AL) inserted on 03/02/2021 -- needed to compute Pi(q,1)
  Prododdresult = Kummer_realcorrection + Kummer_res;  // = log(kummer)
  fprintf(stdout, "log Pi_odd = %s\n", fstr(Prododdresult));
  
  __float128 radq = sqrtq(q);
  MinimumLoddresult = pi * sqrtq(MinimumL)/(q*radq); /* min L for odd characters */
  MaximumLoddresult = pi * sqrtq(MaximumL)/(q*radq); /* max L for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L (1,chi) \n ");
  printf("chiminoddL index = ");
  printf("%ld\n",jminL);
  printf("chimaxoddL index = ");
  printf("%ld\n",jmaxL);
  
  MinimumLprimeoddresult=pi*sqrtq(MinimumLprime)/radq; /* min L' for odd characters */
  MaximumLprimeoddresult=pi*sqrtq(MaximumLprime)/radq; /* max L' for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters L'(1,chi) \n ");
  printf("chiminoddLprime index = ");
  printf("%ld\n",jminLprime);
  printf("chimaxoddLprime index = ");
  printf("%ld\n",jmaxLprime);
  
  Minimumbchioddresult=sqrtq(Minimumbchi); /* min |bchi| for odd characters */
  Maximumbchioddresult=sqrtq(Maximumbchi); /* max |bchi| for odd characters */
  printf("\n *** indexes for minimal and maximal odd characters |b_chi| \n ");
  printf("chiminoddbchi index = ");
  printf("%ld\n",jminbchi);
  printf("chimaxoddbchi index = ");
  printf("%ld\n",jmaxbchi);

   
  myresult.res = oddresult;
  myresult.max = Maximumoddresult;
  myresult.min = Minimumoddresult;  
  myresult.Kummer = Kummer;
  // (AL) inserted on 03/02/2021 -- needed to compute Pi(q,1)
  myresult.H1class = H1class;
  myresult.Prodresult = Prododdresult;
  
  // (AL) inserted on 28/05/2021 for gamma_Kr
  for (ptrdiff_t j = 0; j < NUM_DIV; j++) { myresult.Dirichletclassresult[j] = q * Oddclasssumr[j] ; };

  
  myresult.maxL = MaximumLoddresult;
  myresult.minL = MinimumLoddresult;
  myresult.maxLprime = MaximumLprimeoddresult;
  myresult.minLprime = MinimumLprimeoddresult;
  myresult.maxbchi = Maximumbchioddresult;
  myresult.minbchi = Minimumbchioddresult;
  
  // (AL) inserted on 19/11/2022 -- needed to compute L(1,chi_quad)
  myresult.quadcharvalue = quadcharvalue;


  
  clock_t endcompodd = clock(); /* stop clock */

  printf("\n *** First: oddresult = ");
  { 
    __float128 r = oddresult;
    int width = 46;
    int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
    if ((size_t) nquad < sizeof bufRe)
      printf ("%s", bufRe);
  }
  printf(",  (res = ");
  { 
    __float128 r = res;
    int width = 46;
    int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
    if ((size_t) nquad < sizeof bufRe)
      printf ("%s", bufRe);
  }
  printf(")\n");
  
  
  print_time("OddResult computation", startcompodd, endcompodd);
  
  fflush(stdout);
  
  return myresult;
}

fftwq_complex * mkS(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, int precision, const char * basename, int * fd_S) {
  
  /* STEP: S 
   *  need: fftin array available for writing
   *  provide: S data in fftin to compute <basename>-Sfft.qd
   *  provide: 
   */
  
  clock_t startcomputeS = clock();
  
  /* inserting the null imaginary parts */   
  long int s_a = 1.0Q; 
  __float128 s_aq = 0.0Q;
  for  (long int k = 0; k < nhalf + 2; k++) {
    s_aq = (__float128) ((__float128) s_a / (__float128) q);
    fftin[k][REAL] = s(s_aq, precision, 0); 
    fftin[k][IMAG] = 0.0Q;
    s_a = (s_a * primroot) % q;
  }
  
  clock_t endcomputeS = clock();
  fflush(stdout);
  print_time("S computation", startcomputeS, endcomputeS);

  if (basename == NULL) {
    /* copy of the file not requested */
    return NULL;
  }

  fftwq_complex * S = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_S, 'r', fd_S);

  return S;
}

fftwq_complex * fftS(fftwq_complex * fftin, fftwq_complex * fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_Sfft) {
  
  /* STEP: fftS 
   *  need: fftin mmapped array with S data 
   *  provide: file <basename>-Sfft.qd (by copying fftout)
   */
 
  clock_t startfft  = clock();
  fftwq_execute(plan);
  clock_t endfft = clock();
  print_time("Sfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwq_complex * S_fft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTS, 'r', fd_Sfft);

  return S_fft;
}

fftwq_complex * mkLGeven(fftwq_complex * fftin, ptrdiff_t nhalf, long int q, long int primroot, const char *basename, int * fd_LGeven) {
  
  /*
   * STEP: mkLGeven 
   *  need: fftin mmapped array
   *  provide: file <basename>-LGeven.qd
   *
   */
  
  __float128 pi = 3.1415926535897932384626433832795028841971693993751;
  // __float128 logpi = logq(pi);
  __float128 d;
  long int a;  
  long int k;

  clock_t startcomp = clock();
  a = 1; 
  for (k = 0; k < nhalf; k++){  
     d = (__float128) a/q ;
     fftin[k][REAL]= - logq(sinq(pi*d)); /* logpi is transformed to 0 by orthogonality of Dirichlet characters*/
     /* fftin[k][REAL]= logpi - logq(sinq(pi*d)); */ /* values of log(gamma(a_k/q) via the reflection formula*/
     fftin[k][IMAG] = 0.0Q;  
     a = (a * primroot) % q;
  }
  clock_t endcomp = clock();
  print_time("mkLGeven", startcomp, endcomp);
  fflush(stdout);
  
  if (basename == NULL) {
    /* copy of the file not requested, output in fftin */
    return NULL;
  }

  fftwq_complex * LGeven = (fftwq_complex *) mkmmapcopy(fftin, nhalf, sizeof(fftwq_complex), basename, N_LGEVEN, 'r', fd_LGeven);

  return LGeven;

}

fftwq_complex * fftLGeven(fftwq_complex *fftin, fftwq_complex *fftout, fftwq_plan plan, long int nhalf, const char *basename, int *fd_LGevenfft) {
  
  /* STEP: fftLGeven 
   *  need: fftin mmapped array with LGeven data
   *  provide: file <basename>-LGevenfft.qd (by renaming fftout)
   */
  
  clock_t startfft = clock();
  fftwq_execute(plan);
  clock_t endfft = clock();
  print_time("LGevenfft", startfft, endfft);
  fflush(stdout);

  if (basename == NULL) {
    /* copy of the file not requested, output in fftout */
    return NULL;
  }

  fftwq_complex * LGevenfft = (fftwq_complex *) mkmmapcopy(fftout, nhalf, sizeof(fftwq_complex), basename, N_FFTLGEVEN, 'r', fd_LGevenfft);

  return LGevenfft;

}

struct result evenResult(fftwq_complex *lngamma_evenfft, fftwq_complex *S_fft, ptrdiff_t nhalf, ptrdiff_t q, ptrdiff_t NUM_DIV, ptrdiff_t* divisors) {

  /* STEP: evenResult */

  /* OLD COMMENT FOLLOWS */
  /* Computing EK^+(q) */
  /* Computing min_{chi\ne chi_0} \vert L'/L(1,\chi) \vert =: M(q) */
  /* sum starts from i=1 to avoid the trivial character contribution */
  /* values_trasf contains the DFT of lngamma_even; due to the simmetries the output has half elements than the input */
  /* values_trasf[i][REAL]=values_trasf[nhalf-i][REAL];  values_trasf[i][IMAG] = -values_trasf[nhalf-i][IMAG] */
  /* S_trasf contains the DFT of S; due to the simmetries the output has half elements than the input */
  /* S_trasf[i][REAL]=S_trasf[nhalf-i][REAL];  S_lngamma_even_tras[i][IMAG] = -S_trasf[nhalf-i][IMAG] */
  /*   SO THEY HAVE THE SAME ABSOLUTE VALUES  */

  struct result myresult;
  char bufRe[512];
  
 
  __float128 evenresult = 0.0Q;
  // (AL) inserted 03/02/2021: to compute Pi(q,1)
  __float128 Prod_res = 0.0Q;
  __float128 Prodevenresult = 0.0Q;
    
  __float128 res = 0.0Q;
  __float128 realpart = 0.0Q;
  __float128 imagpart = 0.0Q;
  __float128 Minimumevenresult = 0.0Q;
  __float128 Maximumevenresult = 0.0Q;
  __float128 Minimum = FLT128_MAX; /* even characters */
  __float128 Maximum = 0.0Q; 
  __float128 MinimumLevenresult = 0.0Q;
  __float128 MaximumLevenresult = 0.0Q;
  __float128 MinimumL = FLT128_MAX; /* even characters */
  __float128 MaximumL = 0.0Q; 
  __float128 MinimumLprimeevenresult = 0.0Q;
  __float128 MaximumLprimeevenresult = 0.0Q;
  __float128 MinimumLprime = FLT128_MAX; /* even characters */
  __float128 MaximumLprime = 0.0Q; 
  __float128 Minimumbchievenresult;
  __float128 Maximumbchievenresult;  
  __float128 Minimumbchi = FLT128_MAX;/* even characters */
  __float128 Maximumbchi = 0.0Q;
  
  __float128 lq = (__float128) q;
  __float128 quadcharvalue = 0.0Q;

  long int jmin = 0;
  long int jmax = 0;
  long int jminL = 0;
  long int jmaxL = 0;
  long int jminLprime = 0;
  long int jmaxLprime = 0;
  long int jminbchi = 0;
  long int jmaxbchi = 0;

  __float128 num     = 0.0Q;
  __float128 num1    = 0.0Q;
  __float128 den     = 0.0Q;
  __float128 quot    = 0.0Q;
  __float128 Lquotient = 0.0Q;
  __float128 aux    = 0.0Q;   
  __float128 frac    = 0.0Q;   
  __float128 frac1   = 0.0Q;   
  __float128 absbchi = 0.0Q;
  __float128 bchiREq = 0.0Q;
  __float128 bchiIMq = 0.0Q;  
  __float128 corr    = logq((2.0Q)*pi)+Eulergamma;
  __float128 twicecorr    = (2.0Q) * corr;
    // (AL) inserted on 06/06/2022 -- for computing b_chi  
  __float128 minuslogq    = -logq(q);  

  res = 0.0Q;
  evenresult = 0.0Q;

  // 28/05/2021 for computing the sums over dirichlet character of index ell divided by r
  __float128 *Evenclasssumr;
  Evenclasssumr = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV); 
  for (ptrdiff_t j = 0; j < NUM_DIV; j++) { Evenclasssumr[j] = 0.0Q ;};

  myresult.Dirichletclassresult = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV);
  ptrdiff_t r;
  //
  
  printf("Computing the sum and minmax over even characters ..."); 
  clock_t startcompeven = clock();

  /* sum starts from i=1 to avoid the principal character contribution */ 
     
  for (long int i=1;  i < nhalf; ++i) { 
      num = S_fft[i][REAL] * lngamma_evenfft[i][REAL] +  S_fft[i][IMAG]*lngamma_evenfft[i][IMAG];
      den = lngamma_evenfft[i][REAL] *lngamma_evenfft[i][REAL] + lngamma_evenfft[i][IMAG]*lngamma_evenfft[i][IMAG];
      Lquotient = num / den;
          // 28/05/2021 for gamma_Kr
      Evenclasssumr[0] = Evenclasssumr[0]+ Lquotient;
      res = Evenclasssumr[0];
      for (ptrdiff_t j = 1; j < NUM_DIV; j++) { r = divisors[j];
        					 // r | char index 
           		 if (i%r == 0) {Evenclasssumr[j] = Evenclasssumr[j]+ Lquotient ;}; 
                }; 
               
      // (AL) inserted 03/02/2021: to compute Pi(q,1)
      Prod_res = Prod_res + (0.5Q) * logq(den);
      
      num1 = S_fft[i][IMAG] * lngamma_evenfft[i][REAL] - S_fft[i][REAL]*lngamma_evenfft[i][IMAG];
      realpart = twicecorr * lngamma_evenfft[i][REAL] + S_fft[i][REAL];
      imagpart = twicecorr * lngamma_evenfft[i][IMAG] + S_fft[i][IMAG];
      aux = realpart * realpart + imagpart * imagpart;
      frac = corr + (0.5Q)*Lquotient;
      frac1 = (0.5Q)*num1 / den;
      quot = frac * frac + frac1 * frac1; 
                // 06/06/2022 for bchi
	  bchiREq = -(0.5Q) * Lquotient + minuslogq;
      bchiIMq = frac1; 
      absbchi = bchiREq * bchiREq + bchiIMq * bchiIMq;

      if (quot < Minimum) { Minimum = quot; jmin = 2*i ;};
      if (quot > Maximum) { Maximum = quot; jmax = 2*i ;};
      if (den < MinimumL) { MinimumL = den; jminL = 2*i ;}; 
      if (den > MaximumL) { MaximumL = den; jmaxL = 2*i ;};
      if (aux < MinimumLprime) { MinimumLprime = aux; jminLprime = 2*i ;};
	  if (aux > MaximumLprime) { MaximumLprime = aux; jmaxLprime = 2*i ;};  
	  if (absbchi < Minimumbchi) { Minimumbchi = absbchi; jminbchi = 2*i ;};
      if (absbchi > Maximumbchi) { Maximumbchi = absbchi; jmaxbchi = 2*i ;};
	  

  }   


  if (nhalf % 2 == 0)
  		{ ptrdiff_t quad_index = floor(nhalf/2);
		  quadcharvalue = fabsq((2.0Q) * lngamma_evenfft[quad_index][REAL]/sqrtq(lq));
		};
		  
  Minimumevenresult = sqrtq(Minimum); /* min for even characters */
  Maximumevenresult = sqrtq(Maximum); /* max for even characters */
  printf("\n *** indexes for minimal and maximal even characters L'/L (1,chi) \n ");
  printf("chimineven index = ");
  printf("%ld\n",jmin);
  printf("chimaxeven index = ");
  printf("%ld\n",jmax);
  
  __float128 radq= sqrtq(q);
  MinimumLevenresult = (2.0Q) * sqrtq(MinimumL) / radq; /* min L for even characters */
  MaximumLevenresult = (2.0Q) * sqrtq(MaximumL) / radq; /* max L for even characters */
  printf("\n *** indexes for minimal and maximal even characters L(1,chi) \n ");
  printf("chiminevenL index = ");
  printf("%ld\n",jminL);
  printf("chimaxevenL index = ");
  printf("%ld\n",jmaxL);
  
  MinimumLprimeevenresult = sqrtq(MinimumLprime) / radq; /* min L' for even characters */
  MaximumLprimeevenresult = sqrtq(MaximumLprime) / radq; /* max L' for even characters */
  printf("\n *** indexes for minimal and maximal even characters L'(1,chi) \n ");
  printf("chiminevenLprime index = ");
  printf("%ld\n",jminLprime);
  printf("chimaxevenLprime index = ");
  printf("%ld\n",jmaxLprime);
  
  Minimumbchievenresult=sqrtq(Minimumbchi); /* min |bchi| for even characters */
  Maximumbchievenresult=sqrtq(Maximumbchi); /* max |bchi| for even characters */
  printf("\n *** indexes for minimal and maximal even characters |b_chi| \n ");
  printf("chiminevenbchi index = ");
  printf("%ld\n",jminbchi);
  printf("chimaxevenbchi index = ");
  printf("%ld\n",jmaxbchi);
  
  evenresult = res/(2.0Q); 
  
  // (AL) inserted on 03/02/2021 -- needed to compute Pi(q,1)
  Prodevenresult = (logq(2.0Q) - (0.5Q)*logq( (__float128) q))* (nhalf-(1.0Q)) + Prod_res;
  fprintf(stdout, "log Pi_even = %s\n", fstr(Prodevenresult));
  
  myresult.res = evenresult;
  myresult.max = Maximumevenresult;
  myresult.min = Minimumevenresult; 
  myresult.Kummer = (0.0Q);
  // (AL) inserted on 03/02/2021 -- needed to compute Pi(q,1)
  myresult.H1class = (0.0Q);
  myresult.Prodresult = Prodevenresult;

    // (AL) inserted on 28/05/2021 -- for gamma_Kr
  for (ptrdiff_t j = 0; j < NUM_DIV; j++) { myresult.Dirichletclassresult[j] =  Evenclasssumr[j]/(2.0Q) ; };

  
  myresult.maxL = MaximumLevenresult;
  myresult.minL = MinimumLevenresult;
  myresult.maxLprime = MaximumLprimeevenresult;
  myresult.minLprime = MinimumLprimeevenresult;
  myresult.maxbchi = Maximumbchievenresult;
  myresult.minbchi = Minimumbchievenresult;

  // (AL) inserted on 19/11/2022 -- needed to compute L(1,chi_quad)
  myresult.quadcharvalue = quadcharvalue;
  

  clock_t endcompeven = clock();

 // printf("\n *** Second: evenresult = %32.30Lf, (res = %24.28Lf)\n", evenresult, res);
  printf("\n *** Second: evenresult = ");
  { 
  __float128 r = evenresult;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
 printf(",  (res = ");
  { 
  __float128 r = res;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s", bufRe);
  }
  printf(")\n");
  
  print_time("EvenResult computation", startcompeven, endcompeven);
  
  fflush(stdout);

  return myresult;
}

char bufPrint[512];
void printFloat128(__float128 x) {
  size_t n = quadmath_snprintf(bufPrint, sizeof(bufPrint), "%+-#*.30Qf", 92, x);
  if (n < sizeof(bufPrint))
  printf("%s\n", bufPrint);
  else
  printf("--error--: n = %ld\n", n);
  }

int printMeasure(struct measure m, const char *name, FILE *stream, int evenodd) {

  fprintf(stream, "Measures for %s\n", name);

  fprintf(stream, " NDIFF_2 = Norm2(%s - fft(fft(%s)) = %s\n", name, name, fstr(m.norm2));
  fprintf(stream, " NBASE_2 = Norm2(%s) = %s\n", name, fstr(m.bnorm2));
  fprintf(stream, " Inverse precision for %s in Norm2: NDIFF_2 / NBASE_2 = %s\n", name, fstr(m.prec2));
  fprintf(stream, " NDIFF_1 = Norm1(%s - fft(fft(%s)) = %s\n", name, name, fstr(m.norm1));
  fprintf(stream, " NBASE_1 = Norm1(%s) = %s\n", name, fstr(m.bnorm1));
  fprintf(stream, " Inverse precision for %s in Norm1: NDIFF_1 / NBASE_1 = %s\n", name, fstr(m.prec1));
  fprintf(stream, " NDIFF_0 = Norm00(%s - fft(fft(%s)) = %s\n", name, name, fstr(m.norm00));
  fprintf(stream, " NBASE_0 = Norm00(%s) = %s\n", name, fstr(m.bnorm00));
  fprintf(stream, " Inverse precision for %s in Norm0: NDIFF_0 / NBASE_0 = %s\n", name, fstr(m.prec00));

  if (evenodd) {
    fprintf(stream, " NBASE_2_even = Norm2(%s_even)  = %s\n", name, fstr(m.bnorm2even));
    fprintf(stream, " NBASE_2_odd  = Norm2(%s_odd)   = %s\n", name, fstr(m.bnorm2odd));
    fprintf(stream, " NBASE_1_even = Norm1(%s_even)  = %s\n", name, fstr(m.bnorm1even));
    fprintf(stream, " NBASE_1_odd  = Norm1(%s_odd)   = %s\n", name, fstr(m.bnorm1odd));
    fprintf(stream, " NBASE_0_even = Norm00(%s_odd)  = %s\n", name, fstr(m.bnorm00even));
    fprintf(stream, " NBASE_0_odd  = Norm00(%s_even) = %s\n", name, fstr(m.bnorm00odd));
  }
  
  fprintf(stream, "\n");
  
  return 1;
}


int printNorms(struct norms norm) {
  /* print norm result */
  printMeasure(norm.norm_s, "S", stdout, NO);
  printMeasure(norm.norm_ak, "AK", stdout, NO);
  printMeasure(norm.norm_lgodd, "LGodd", stdout, NO);
  printMeasure(norm.norm_lgeven, "LGeven", stdout, NO);
  return 1;
}



int printResults(long int q, ptrdiff_t nhalf, struct result oddresult, struct result evenresult, ptrdiff_t NUM_DIV, ptrdiff_t* divisors ) {

  char bufRe[512];

  long int qminusone = q-1;
  //
  __float128 *gammaK;
  gammaK = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV);
  __float128 *gammaKplus;
  gammaKplus = (__float128 *) fftwq_malloc(sizeof(__float128) * NUM_DIV);
  
  ptrdiff_t r;
  //
  
  /** RESULTS **/
  printf("*** RESULTS:\n\n");
  fflush(stdout);
  __float128 lg2pi = logq(2*pi);
  __float128 Eulerpluslog2pi = Eulergamma + lg2pi;
 
    // (AL) inserted on 28/05/2021 -- results by divisors
  printf("%s\n", "--------------");
  printf("%s\n", "gamma_Kr constants for every intermediate field");
  
  FILE * gammaKrfile;
  const char* name="gammaKrResults-fftwq-";
  char namefile[50];
  sprintf(namefile,"%s%td.csv",name,q);
  gammaKrfile = fopen (namefile,"a");
  //gammaKrfile = fopen ("gammaKrfile.csv","a");
  //fprintf (gammaKrfile, "q;r;gammaKr(q);gammaKrplus(q);difference\n");
    		
  long int l;  
    
  for (ptrdiff_t j = 0; j < NUM_DIV; j++) {r= divisors[j];
  			if (r%2 == 1) {gammaK[j] = Eulergamma + Eulerpluslog2pi*(qminusone/r-1.0Q)+ oddresult.Dirichletclassresult[j] + evenresult.Dirichletclassresult[j]; }
  			else { l = binarySearch(divisors, 0, NUM_DIV-1, r/2) ; gammaK[j] = gammaKplus[l]; };	
  			if (j < NUM_DIV-1) {gammaKplus[j] = Eulergamma + Eulerpluslog2pi *(qminusone/((2.0Q)*r)-1.0Q) + evenresult.Dirichletclassresult[j]; }
  			else {gammaKplus[j] = Eulergamma ; }; 
  			printf("%s\n", "----");
  			printf("%s", "gammaK_");
  			printf("%tu",r);
  			printf("%s", "("); 
   		    printf("%td",q);
     		printf("%s", ") = "); 
    		//printf("%32.30Lf\n", gammaK[j]);   
    		{
    		__float128 r = gammaK[j];
    		int width = 46;
    		int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
    		if ((size_t) nquad < sizeof bufRe)
    		printf ("%s\n", bufRe);
    		} 
  			printf("%s", "gammaKplus_");
  			printf("%tu",r);
  			printf("%s", "("); 
   		    printf("%td",q);
    		printf("%s", ") = "); 
    		//printf("%32.30Lf\n", gammaKplus[j]); 
    		{
    		__float128 r = gammaKplus[j];
    		int width = 46;
    		int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
    		if ((size_t) nquad < sizeof bufRe)
    		printf ("%s\n", bufRe);
    		}    		
    		fprintf (gammaKrfile, "%td;", q);
    		fprintf (gammaKrfile, "%td;", r);
    		//fprintf (gammaKrfile, "%32.30Lf;", gammaK[j]); 
    		{
    		__float128 r = gammaK[j];
    		int width = 46;
    		int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
    		if ((size_t) nquad < sizeof bufRe)
    		fprintf (gammaKrfile,"%s;", bufRe);
    		}

    		//fprintf (gammaKrfile, "%32.30Lf;", gammaKplus[j]); 
    		{
    		__float128 r = gammaKplus[j];
    		int width = 46;
    		int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
    		if ((size_t) nquad < sizeof bufRe)
    		fprintf (gammaKrfile,"%s;", bufRe);
    		}
    		
    		//fprintf (gammaKrfile, "%32.30Lf", gammaK[j]-gammaKplus[j]); 
    		{
    		__float128 r = gammaK[j]-gammaKplus[j];
    		int width = 46;
    		int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
    		if ((size_t) nquad < sizeof bufRe)
    		fprintf (gammaKrfile,"%s", bufRe);
    		}
    		
    		fprintf (gammaKrfile, "\n");

			};
  fclose (gammaKrfile);			
  printf("%s\n", "--------------");
  //
  printf("%s\n", "Other quantities");

  __float128 EK = gammaK[0];
  __float128 gammaK2 = gammaKplus[0];
  __float128 EKdiff = oddresult.res + Eulerpluslog2pi * nhalf;

  
  // (AL) inserted on 03/02/2021 per il calcolo di Pi(q,1)
  __float128 Prod = expq(- oddresult.Prodresult - evenresult.Prodresult);
  __float128 residue = expq(oddresult.Prodresult + evenresult.Prodresult);
  
  if (EK < 0 || gammaK2 < 0 ) {
    FILE * negativeEKfile;
    negativeEKfile = fopen ("../negativeEK.txt","a");
    fprintf (negativeEKfile, "q = %td, ", q);
    //fprintf (negativeEKfile, "EK = %32.30Lf, ", EK); 
    fprintf(negativeEKfile, "EK = "); 
    {
     __float128 r = EK;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile,"%s", bufRe);
    }
    fprintf(negativeEKfile, ", ");
    
    //fprintf (negativeEKfile, "EK+ = %32.30Lf, ", gammaK2); 
    fprintf(negativeEKfile, "EK+ = "); 
    {
     __float128 r = gammaK2;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile, "%s", bufRe);
    }
    fprintf(negativeEKfile, ", ");
    
    //fprintf (negativeEKfile, "EK-EK+ = %32.30Lf", EKdiff); 
     fprintf(negativeEKfile, "EK-EK+ = "); 
     {
     __float128 r = EKdiff;
     int width = 46;
     int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%+-#*.30Qf", width, r);
     if ((size_t) nquad < sizeof bufRe)
     fprintf (negativeEKfile,"%s", bufRe);
    } 
    fprintf (negativeEKfile, "\n");
    fclose (negativeEKfile);
  };

  printf("%s\n", "--------------");
  /* print EK_q */
  printf("%s", "[1] EK(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", EK);
  {
  __float128 r = EK;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  /* print EK_q^+ */
  printf("%s", "[2] EKplus(");
  printf("%td",q);
  printf("%s", ") = ");
 //printf("%32.30Lf\n", EK_plus); 
  {
  __float128 r = gammaK2;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  /* print EK_q- EK_q^+ */
  printf("%s", "[3] EK(");
  printf("%td",q);
  printf("%s", ")diff = "); 
  //printf("%32.30Lf\n", EKdiff); 
  {
  __float128 r = EKdiff;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }
  
  

  printf("%s\n", "--------------");

  
  /* print min_q */
  printf("%s", "[4] min_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.min); 
 {
  __float128 r = evenresult.min;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }


  /* print min_q */
  printf("%s", "[5] min_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.min);  
  {
  __float128 r = oddresult.min;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }

  
  /* print min_q */
  printf("%s", "[6] min_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fminq(evenresult.min, oddresult.min));  
  {
  __float128 r = fminq(evenresult.min, oddresult.min);
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }  
  
  printf("%s\n", "--------------");
  
  /* print max_q */
  printf("%s", "[7] max_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.max); 
  {
  __float128 r = evenresult.max;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }    

  /* print max_q */
  printf("%s", "[8] max_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.max);  
  {
  __float128 r = oddresult.max;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  } 
  
  /* print min_q */
  printf("%s", "[9] max_{chi neq chi_0}|L'/L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fmaxq(evenresult.max, oddresult.max));   
  {
  __float128 r = fmaxq(evenresult.max, oddresult.max);
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }   
  
  printf("%s\n", "--------------");  
  
  /* print Kummer ratio */
  printf("%s", "[10] Kummer_ratio(");
  printf("%td",q);
  printf("%s", ") = ");
  /*printf("%Lf\n", Kummer);*/
  //printf("%32.30Lf\n", oddresult.Kummer);   
  {
  __float128 r = oddresult.Kummer;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }   
  
  
  // (AL) inserted on 03/02/2021 per il calcolo di class number e Pi(q,1)
  
  printf("%s", "[11] log(Kummer) = log r(");
  printf("%td",q);
  printf("%s", ") = ");
  /*printf("%Lf\n", logq(Kummer));*/
  //printf("%32.30Lf\n", oddresult.Prodresult); //logq(oddresult.Kummer= oddresult.Prodresult
  {
  __float128 r = oddresult.Prodresult;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }  
  
  /* print log h1 */
  printf("%s", "[12] logh_1(");
  printf("%td",q);
  printf("%s", ") = ");
  /*printf("%Lf\n", H1class);*/
  //printf("%32.30Lf\n", oddresult.H1class); 
  {
  __float128 r = oddresult.H1class;
   int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
   printf ("%s\n", bufRe);
  }     
  
  printf("%s\n", "--------------");  
  
  /* print Pi(q,1) */
  printf("%s", "[13] Pi(");
  printf("%td",q);
  printf("%s", ",1.0Q) = ");
  /*printf("%Lf\n", expq(Prododdresult + Prodevenresult));*/
  //printf("%32.30Lf\n", Prod );   
  {
  __float128 r = Prod;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }  
  
  /* print serie di Laurent */
  printf("%s", "[14] Laurent series of zeta_{Q(zeta_q)} at 1: c_{-1} = Pi(q,1)^(-1) = ");
  //printf("%32.30Lf\n", residue );  
  {
  __float128 r = residue;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  
  printf("%s", "[15] c_{0} = c_{-1} * EK = ");
  /*printf("%Lf\n", expq(Prododdresult + Prodevenresult));*/
  //printf("%32.30Lf\n",  EK * residue );  
  {
  __float128 r = EK * residue;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    

  
  printf("%s\n", "--------------");
  
    /* print minL_q */
  printf("%s", "[16] minL_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.minL); 
  {
  __float128 r = evenresult.minL;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  

  /* print minL_q */
  printf("%s", "[17] minL_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.minL);  
  {
  __float128 r = oddresult.minL;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  

  
  /* print minL_q */
  printf("%s", "[18] minL_{chi neq chi_0}|L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fminq(evenresult.minL, oddresult.minL));   
  {
  __float128 r = fminq(evenresult.minL, oddresult.minL);
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
   
  
  printf("%s\n", "--------------");
  
  /* print maxL_q */
  printf("%s", "[19] maxL_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.maxL); 
  {
  __float128 r = evenresult.maxL;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  

  /* print maxL_q */
  printf("%s", "[20] maxL_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.maxL);  
  {
  __float128 r = oddresult.maxL;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    

  
  /* print maxL_q */
  printf("%s", "[21] maxL_{chi neq chi_0}|L(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fmaxq(evenresult.maxL, oddresult.maxL));   
  {
  __float128 r = fmaxq(evenresult.maxL, oddresult.maxL);
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  
  
  printf("%s\n", "--------------");   
  
  /* print minLprime_q */
  printf("%s", "[22] minLprime_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.minLprime); 
  {
  __float128 r = evenresult.minLprime;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    

  /* print minLprime_q */
  printf("%s", "[23] minLprime_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.minLprime);  
  {
  __float128 r = oddresult.minLprime;
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  
  /* print minLprime_q */
  printf("%s", "[24] minLprime_{chi neq chi_0}|Lprime(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fminq(evenresult.minLprime, oddresult.minLprime));    
  {
  __float128 r = fminq(evenresult.minLprime, oddresult.minLprime);    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }  
  
  printf("%s\n", "--------------");
  
  /* print maxLprime_q */
  printf("%s", "[25] maxLprime_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.maxLprime); 
  {
  __float128 r = evenresult.maxLprime;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }   

  /* print maxLprime_q */
  printf("%s", "[26] maxLprime_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.maxLprime);  
  {
  __float128 r = oddresult.maxLprime;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    

  
  /* print maxLprime_q */
  printf("%s", "[27] maxLprime_{chi neq chi_0}|Lprime(1,chi)|(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fmaxq(evenresult.maxLprime, oddresult.maxLprime));   
  {
  __float128 r = fmaxq(evenresult.maxLprime, oddresult.maxLprime);   
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }  
    
  printf("%s\n", "--------------");   
  
   /* print minbchi_q */
  printf("%s", "[28] minbchi_even(");
  printf("%td",q);
  printf("%s", ") = ");
 // printf("%32.30Lf\n", evenresult.minbchi); 
   {
  __float128 r = evenresult.minbchi;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }   


  /* print minbchi_q */
  printf("%s", "[29] minbchi_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.minbchi);  
  {
  __float128 r = oddresult.minbchi;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }   


  
  /* print minbchi_q */
  printf("%s", "[30] minbchi_{chi neq chi_0}(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fminl(evenresult.minbchi, oddresult.minbchi));    
  {
  __float128 r = fminq(evenresult.minbchi, oddresult.minbchi);    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }  
  
  printf("%s\n", "--------------");
  
  /* print maxbchi_q */
  printf("%s", "[31] maxbchi_even(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", evenresult.maxbchi); 
  {
  __float128 r = evenresult.maxbchi;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    
  

  /* print maxbchi_q */
  printf("%s", "[32] maxbchi_odd(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", oddresult.maxbchi);  
  {
  __float128 r = oddresult.maxbchi;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }    

  
  /* print minbchi_q */
  printf("%s", "[33] maxbchi_{chi neq chi_0}(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", fmaxl(evenresult.maxbchi, oddresult.maxbchi));   
  {
  __float128 r = fmaxq(evenresult.maxbchi, oddresult.maxbchi);    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }
  
  printf("%s\n", "--------------"); 
  
  /* print L(q,chi_quad) */  
   __float128 quadcharvalue = 0;
  if (nhalf % 2 == 0) 
  		{
  			printf("%s\n", "******** the quadratic character is even");
  			quadcharvalue = evenresult.quadcharvalue;
  		}
  else
    	{
    		printf("%s\n", "******** the quadratic character is odd");
	    	quadcharvalue = oddresult.quadcharvalue;
	    };
	    
  printf("%s", "[34] L(1,chi_quad)(");
  printf("%td",q);
  printf("%s", ") = ");    	
  //printf("%32.30Lf\n", quadcharvalue);  

  {
  __float128 r = quadcharvalue;    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }


  printf("%s", "[35] L(1,chi_quad)/log(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", quadcharvalue/logl(q)); 
  
  {
  __float128 r = quadcharvalue/logq(q);    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }
   
  printf("%s", "[36] L(1,chi_quad)/loglog(");
  printf("%td",q);
  printf("%s", ") = ");
  //printf("%32.30Lf\n", quadcharvalue/logl(logl(q)));  

  {
  __float128 r = quadcharvalue/logq(logq(q));    
  int width = 46;
  int nquad = quadmath_snprintf (bufRe, sizeof bufRe, "%-#*.30Qf", width, r);
  if ((size_t) nquad < sizeof bufRe)
  printf ("%s\n", bufRe);
  }

  printf("%s\n", "--------------");    
  
  return 1;
}

void apriori(long int q, int precision, FILE * stream) {

  __float128 N1LGeven = 0.0Q;
  __float128 N00LGeven = 0.0Q;
  /* __float128 N1LGodd = 0.0Q; */
  __float128 N00LGodd = 0.0Q; /**/
  __float128 Z0 = -2.00635645590858485121010002;
  __float128 N1Seven = 0.0Q;
  __float128 N00Seven = 0.0Q;
  __float128 loglq = logq((__float128) q);
  __float128 lq = (__float128) q;
  __float128 N1AKodd = 0.0Q;
  __float128 N00AKodd = 0.0Q;
  __float128 N2AKodd = 0.0Q;
    
  N1LGeven = (0.5Q) * ((q - 1.0Q) * logq((2.0Q)) - loglq);
  N00LGeven = - logq(sinq(pi / lq));
  N1Seven = - Z0 * (q - 1.0Q) - loglq * logq((2.0Q) * pi) - 0.5 * (loglq * loglq);
  N00Seven = - (2.0Q) * s(1.0 / lq, precision, 0) + s(1.0 - 1.0/lq, precision, 0);
  N1AKodd = (q - 1.0Q) * 0.25 * (q - 1.0Q) / lq;
  N00AKodd = (q - 2) / lq;
  // inserted AleL 13/03/2021
  N2AKodd = sqrtq((q- 2.0Q)*(q - 1.0Q) / ((6.0Q)*q));
  /* N1LGodd = -1; */
  N00LGodd = 2.0 * lgammaq(1.0 / lq) + logq(sin(pi / lq) / pi);

  fprintf(stream, "*** A priori (closed formulas) norms ***\n");
  
  fprintf(stream, "N1LGeven  = %s\n", fstr(N1LGeven));
  fprintf(stream, "N00LGeven = %s\n", fstr(N00LGeven));
  /* NA fprintf(stream, "N1LGodd   = %s\n", fstr(N1LGodd)); */
  fprintf(stream, "N00LGodd  = %s\n", fstr(N00LGodd));
  fprintf(stream, "N1Seven   = %s\n", fstr(N1Seven));
  fprintf(stream, "N00Seven  = %s\n", fstr(N00Seven));
  fprintf(stream, "N1AKodd   = %s\n", fstr(N1AKodd));
  fprintf(stream, "N00AKodd  = %s\n", fstr(N00AKodd));
  fprintf(stream, "N2AKodd   = %s\n", fstr(N2AKodd));
}


int dostep(const char * step, const char * steps) {
  if (strcasestr(steps, step) != NULL) {
    return 1;
  }
  return 0;
}

int runComputationq(long int q, long int primroot, int precision, const char * basename, const char * steps) {

  ptrdiff_t NUM_POINTS = q-1;
  ptrdiff_t nhalf = NUM_POINTS / 2;

  int fd_fftin;
  int fd_fftout;
  // int fd_fftinverse;
  int fd_LGodd;
  int fd_LGoddfft;
  int fd_AK;
  int fd_AKfft;
  int fd_S;
  int fd_Sfft;
  int fd_LGeven;
  int fd_LGevenfft;
  
  fftwq_complex * fftin  = NULL; // mkfftin(nhalf, basename, &fd_fftin);
  fftwq_complex * fftout = NULL; // mkfftout(nhalf, basename, &fd_fftout);
  fftwq_complex * fftinverse;    // mkfftinverse(nhalf, basename, &fd_fftinverse);
  fftwq_complex * lgodd = NULL;
  fftwq_complex * lgoddfft = NULL;
  fftwq_complex * ak = NULL;
  fftwq_complex * akfft = NULL;
  fftwq_complex * S = NULL;
  fftwq_complex * Sfft = NULL;
  fftwq_complex * lgeven = NULL;
  fftwq_complex * lgevenfft = NULL;
  
  fftwq_plan plan = NULL; 
  fftwq_plan inverse_plan = NULL; 

  struct result oddresult;
  struct result evenresult;
  struct norms allnorms;
  
  int doall = 0; /* make every step */
  int docompute = 0; /* make compute step */
  
  fprintf(stdout, "runComputation: steps: %s\n", steps);
  // (AL) inserted on 28/05/2021 -- needed to get the divisors of (q-1)/2
  FILE *myFile;
  myFile = fopen("divfile.res", "r");
    
  if (myFile == NULL){
    printf("\n ******  ERROR: File divfile.res doesn't exist\n");
    exit (0);
  }
  
  /*getting the number of divisors of (q-1)/2 */
  ptrdiff_t NUM_DIV;
  if (fscanf(myFile, "%tu", &NUM_DIV) != 1) {
    fprintf(stderr, "runComputation(): WARNING unable to read 1 item from \"divfile.res\"\n");
  }
fclose(myFile);
  
  ptrdiff_t *divisors;
  divisors = (ptrdiff_t *) fftwq_malloc(sizeof(ptrdiff_t) * NUM_DIV);
  
  printf("\n ****** GETTING DIVISORS FROM FILE \n");
  getDivisors(divisors,NUM_DIV);
  
  //stampaint(divisors,NUM_DIV);  
  
  if (dostep(ALL, steps)) {
    doall = 1;
  }

  if (dostep(COMPUTE, steps)) {
    docompute = 1;
  }
  
  fftin  = mkfftin(nhalf, basename, &fd_fftin);
  fftout = mkfftout(nhalf, basename, &fd_fftout);
  // fftinverse = mkfftinverse(nhalf, basename, &fd_fftinverse);
  
  if (fftin == NULL || fftout == NULL)
    return 0;
  
  if (doall || dostep(COMPUTE, steps)) {
  
    plan = mkplan(fftin, fftout, nhalf);
    if (plan == NULL)
      return 0;

  }

  if (doall || docompute || dostep(ODDRESULT, steps)) {

    fprintf(stdout, "COMPUTE ODD\n");
    
    if (doall || docompute || dostep(MKLGODD, steps)) {
      lgodd     = mkLGodd(fftin, nhalf, q, primroot, basename, &fd_LGodd);       /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_LGODD);
      lgodd     = mapfile(N_LGODD, basename, nhalf, &fd_LGodd);
    }
    
    if (doall || docompute || dostep(FFTLGODD, steps)) {
      lgoddfft  = fftLGodd(fftin, fftout, plan, nhalf, basename, &fd_LGoddfft);
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_FFTLGODD);
      lgoddfft  = mapfile(N_FFTLGODD, basename, nhalf, &fd_LGoddfft);
    }
    
    if (doall || docompute || dostep(MKAK, steps)) {
      ak        = mkAK(fftin, nhalf, q, primroot, basename, &fd_AK);             /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_AK);
      ak        = mapfile(N_AK, basename, nhalf, &fd_AK);
    }
    
    if (doall || docompute || dostep(FFTAK, steps)) {
      akfft     = fftAK(fftin, fftout, plan, nhalf, basename, &fd_AKfft);        /* qui non servirebbe salvare */
    } else {
      fprintf(stdout, " - Mapping %s in memory\n", N_FFTAK);
      akfft     = mapfile(N_FFTAK, basename, nhalf, &fd_AKfft);
    }
    
    oddresult   = oddResult(lgoddfft, akfft, nhalf, q, NUM_DIV, divisors);

    writeResult(oddresult, basename, N_ODDRESULT);
    
    unmap_File2Array(lgodd, nhalf, fd_LGodd);
    unmap_File2Array(lgoddfft, nhalf, fd_LGoddfft);
    unmap_File2Array(ak, nhalf, fd_AK);
    unmap_File2Array(akfft, nhalf, fd_AKfft);

  } else {

    oddresult = readResult(basename, N_ODDRESULT);

  }
  
  if (doall || docompute || dostep(EVENRESULT, steps)) {
  
    fprintf(stdout, "COMPUTE EVEN\n");

    if (doall || docompute || dostep(MKS, steps)) {
      S         = mkS(fftin, nhalf, q, primroot, precision, basename, &fd_S);    /* qui non servirebbe salvare */
    } else {
      S         = mapfile(N_S, basename, nhalf, &fd_S);
    }
    
    if (doall || docompute || dostep(FFTS, steps)) {
      Sfft      = fftS(fftin, fftout, plan, nhalf, basename, &fd_Sfft);
    } else {
      Sfft      = mapfile(N_FFTS, basename, nhalf, &fd_Sfft);
    }
    
    if (doall || docompute || dostep(MKLGEVEN, steps)) {
      lgeven    = mkLGeven(fftin, nhalf, q, primroot, basename, &fd_LGeven);      /* qui non servirebbe salvare */
    } else {
      lgeven    = mapfile(N_LGEVEN, basename, nhalf, &fd_LGeven);
    }
    
    if (doall || docompute || dostep(FFTLGEVEN, steps)) {
      lgevenfft = fftLGeven(fftin, fftout, plan, nhalf, basename, &fd_LGevenfft); /* qui non servirebbe salvare */
    } else {
      lgevenfft = mapfile(N_FFTLGEVEN, basename, nhalf, &fd_LGevenfft);
    }
    
    evenresult  = evenResult(lgevenfft, Sfft, nhalf, q, NUM_DIV, divisors);

    writeResult(evenresult, basename, N_EVENRESULT);
    
    unmap_File2Array(S, nhalf, fd_S);
    unmap_File2Array(Sfft, nhalf, fd_Sfft);
    unmap_File2Array(lgeven, nhalf, fd_LGeven);
    unmap_File2Array(lgevenfft, nhalf, fd_LGevenfft);
    
  } else {

    evenresult = readResult(basename, N_EVENRESULT);

  }

  /* Destroy the forward plan */
  if (plan != NULL)
    fftwq_destroy_plan(plan);

  if (fftin != NULL) 
    unmap_File2Array(fftin, nhalf, fd_fftin);

  if (fftout != NULL)
    unmap_File2Array(fftout, nhalf, fd_fftout);

  
  /*  From here if all the computation are availables
   *  we compute the inverse of the FFTs and the
   *  the norms of the differences: V - FFT(FFT(V))
   *  for the various input vectors. All mapped
   *  file should be closed
   */

  if (doall
      || dostep(INV, steps)
      || dostep(ILGODD, steps) 
      || dostep(IAK, steps) 
      || dostep(IS, steps) 
      || dostep(ILGEVEN, steps)) {

    fftin  = mkfftin(nhalf, basename, &fd_fftin);
    fftout = mkfftout(nhalf, basename, &fd_fftout);
    // fftinverse = mkfftinverse(nhalf, basename, &fd_fftinverse);
    if (fftin == NULL || fftout == NULL)
      return 0;
 
    inverse_plan = mkinverseplan(fftin, fftout, nhalf);
    if (inverse_plan == NULL)
      return 0;

  }

  if (doall || dostep(INV, steps) || dostep(ILGODD, steps)) {
    fprintf(stdout, "INVERSE LGodd\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTLGODD,  N_FFTLGODD_INV,  &fdinverse);
    if (fftinverse == NULL) {
      fprintf(stderr, "runComputation() fftinverse (LGOdd) == NULL\n");
    }
  }
  
  if (doall || dostep(INV, steps) || dostep(IAK, steps)) {
    fprintf(stdout, "INVERSE AK\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTAK,     N_FFTAK_INV,     &fdinverse);
    if (fftinverse == NULL) {
      fprintf(stderr, "runComputation() fftinverse (AK) == NULL\n");
    }
  }
  
  if (doall || dostep(INV, steps) || dostep(IS, steps)) {
    fprintf(stdout, "INVERSE S\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTS,      N_FFTS_INV,      &fdinverse);
    if (fftinverse == NULL) {
      fprintf(stderr, "runComputation() fftinverse (S) == NULL\n");
    }
  }

  if (doall || dostep(INV, steps) || dostep(ILGEVEN, steps)) {
    fprintf(stdout, "INVERSE LGeven\n");
    int fdinverse;
    fftinverse = fft_INV(fftin, fftout, inverse_plan, nhalf, basename, N_FFTLGEVEN, N_FFTLGEVEN_INV, &fdinverse);
    if (fftinverse == NULL) {
      fprintf(stderr, "runComputation() fftinverse (LGEven) == NULL\n");
    }
  }

  if (inverse_plan == NULL) {
    fftwq_destroy_plan(inverse_plan);
  }

  if (fftin != NULL) 
    unmap_File2Array(fftin, nhalf, fd_fftin);

  if (fftout != NULL)
    unmap_File2Array(fftout, nhalf, fd_fftout);

  fftwq_cleanup();

  if (doall || dostep(NORMS, steps) || dostep(SNORM, steps)) {
    fprintf(stdout, "\nNORM OF S ... \n");
    computenormskahan(basename, N_S,      N_FFTS_INV,      nhalf, &(allnorms.norm_s), YES);
    printMeasure(allnorms.norm_s, "S", stdout, NO);
  }
    
  if (doall || dostep(NORMS, steps) || dostep(AKNORM, steps)) {
    fprintf(stdout, "\nNORM OF AK ... \n");
    computenormskahan(basename, N_AK,     N_FFTAK_INV,     nhalf, &(allnorms.norm_ak), YES);
    printMeasure(allnorms.norm_ak, "AK", stdout, NO);
  }

  if (doall || dostep(NORMS, steps) || dostep(LGODDNORM, steps)) {
    fprintf(stdout, "\nNORM OF LGODD ... \n");
    computenormskahan(basename, N_LGODD,  N_FFTLGODD_INV,  nhalf, &(allnorms.norm_lgodd), NO);
    printMeasure(allnorms.norm_lgodd, "LGodd", stdout, NO);
  }
  
  if (doall || dostep(NORMS, steps) || dostep(LGEVENNORM, steps)) {
    fprintf(stdout, "\nNORM OF LGEVEN ... \n");
    computenormskahan(basename, N_LGEVEN, N_FFTLGEVEN_INV, nhalf, &(allnorms.norm_lgeven), NO);
    printMeasure(allnorms.norm_lgeven, "LGeven", stdout, NO);
  }

  printResults(q, nhalf, oddresult, evenresult, NUM_DIV, divisors);

  apriori(q, precision, stdout);

  return 0;
}

