#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include "type.h"
#include "table.h"
#include "computeq.h"

/* from mmap.c */
void  print_time(const char * message, double startplan, double endplan);

#define ARRAYSIZE 8192

void usage(char *command) {
  fprintf(stderr, "USAGE: %s q g precision basename [step[,step[,step[...]]]]\n", command);
  fprintf(stderr, "     where: <q> is a prime number\n"); 
  fprintf(stderr, "            <g> is the output of EKmain (primroot.res)\n");
  fprintf(stderr, "            <precision> is the needed precision digits\n");
  fprintf(stderr, "            <basename> is the radix name of the various output and temporary files \n");
  fprintf(stderr, "            <step,step,....,step> list of steps to perform, Use \"%s --describe\" for explanations)\n", command);
}

void describe() {
  printf("List of steps that can be requested\n");
  printf(" All the following: ALL\n");
  printf("%s: compute oddresult \n", ODDRESULT);
  printf("  %s: compute the values of lngamma [(odd) decimated in frequencies] Output: basename%s\n", MKLGODD, N_LGODD);
  printf("  %s: compute the fft. Input: basename%s, Output: basename%s\n", FFTLGODD, N_LGODD, N_FFTLGODD);
  printf("  %s: compute the values of akvalues [(odd) decimated in frequencies] Output: basename%s\n", MKAK, N_AK);
  printf("  %s: compute the fft. Input: basename%s, Output: basename%s\n", FFTAK, N_AK, N_FFTAK);
  printf("%s: compute evenresult \n", EVENRESULT);
  printf("  %s: compute the S values [(even) decimated in frequencies]. Output: basename-%s\n", MKS, N_S);
  printf("  %s: compute the fft of S. Input: basename%s, Output: basename%s\n", FFTS, N_S, N_FFTS);
  printf("  %s: compute the values of lngamma [(even) decimated in frequencies] Output: basename%s\n", MKLGEVEN, N_LGEVEN);
  printf("  %s: compute the fft. Input: basename%s, Output: basename%s\n", FFTLGEVEN, N_LGEVEN, N_FFTLGEVEN);
  printf("%s: compute all steps \n", ALL);
  printf("%s: perform all computation steps and stops (%s, %s)\n", COMPUTE, ODDRESULT, EVENRESULT);
  printf("%s: compute the inverse of the FFTs (prerequisite: %s) \n", INV, COMPUTE);
  printf("%s: compute the inverse of the LGODD  FFT (prerequisite: %s) \n", ILGODD, COMPUTE);
  printf("%s: compute the inverse of the AK     FFT (prerequisite: %s) \n", IAK, COMPUTE);
  printf("%s: compute the inverse of the S      FFT (prerequisite: %s) \n", IS, COMPUTE);
  printf("%s: compute the inverse of the LGEVEN FFT (prerequisite: %s) \n", ILGEVEN, COMPUTE);
  printf("%s: compute all the norms (prerequisite: %s, %s)\n", NORMS, COMPUTE, INV);
}

int main(int argc, char *argv[]) {

  int precision = 0;
  long int q = 0.0;
  long int g = 0.0;
  char * basename = NULL;
  char * steps = NULL;

  /* Timing */
  clock_t runBegin;
  clock_t runEnd;
  
  if ((argc != 6) && (argc != 2)) {
    usage(argv[0]);
    exit(1);
  }

  if (argc == 2) {
    if (strncmp(argv[1], "--describe", strlen("--describe")) == 0) {
      describe();
      exit(0);
    } else {
      usage(argv[0]);
      exit(1);
    }
  }
  
  runBegin = clock();

  q         = atol(argv[1]);
  g         = atol(argv[2]);
  precision = atoi(argv[3]);
  basename  = argv[4];
  steps     = argv[5];

  printf("*****  Quadruple precision version: precision forced to 128 bits\n");
  precision = 128; /* quadruple precision version */
  fprintf(stdout, "main steps: %s\n", steps);

  
  runComputationq(q, g, precision, basename, steps);
  
  runEnd = clock();

  print_time("TOTAL TIME", runBegin, runEnd);
  
  exit(0);
  
}
