#ifndef TYPE
#define TYPE

#include <math.h>

/* math.c: lngamma() */
__float128 Ld_lgamma(__float128 x);
long double ld_lgamma(long double x);
double d_lgamma(double x);
float f_lgamma(float x);

#define max(a,b) \
   ({ __typeof__ (a) _a = (a); \
       __typeof__ (b) _b = (b); \
     _a > _b ? _a : _b; })

#define FBUFSIZE 128
#define QUADWIDTH 42
#define LDWIDTH 42
#define DWIDTH 30
#define FWIDTH 20
#define LOGPI 1.1447298858494001741434273513530587117
#define LOG2 0.69314718055994530941723212145817656807
#define LOG10 2.3025850929940456840179914546843642076
#define GAMMAX2 1.1544313298030657212130241801648048621
// char fbuf[FBUFSIZE];


#ifdef F128
#include <quadmath.h>
#define __USE_GNU 1
#define MYFLOAT __float128
#define LOG(X) logq(X)
#define CEIL(X) ceilq(X)
#define EXP(X) expq(X)
#define SIN(X) sinq(X)
#define ABS(X) fabsq(X)
#define LGAMMA(X) Ld_lgamma(X)
#define PIGRECO M_PIq
#endif


#ifdef F80
#define __USE_GNU 1
#define MYFLOAT long double
#define LOG(X) logl(X)
#define CEIL(X) ceill(X)
#define EXP(X) expl(X)
#define SIN(X) sinl(X)
#define ABS(X) fabsl(X)
#define LGAMMA(X) ld_lgamma(X)
#define PIGRECO 3.141592653589793238462643383279502884L
#endif

#ifdef F64
#define __USE_GNU
#define __HAVE_FLOAT64 1
#define MYFLOAT double
#define LOG(X) log(X)
#define CEIL(X) ceil(X)
#define EXP(X) exp(X)
#define SIN(X) sin(X)
#define ABS(X) fabs(X)
#define LNGAMMA(X) gsl_sf_lngamma(X);
#define LGAMMA(X) d_lgamma(X)
#define PIGRECO 3.141592653589793238462643383279502884
#endif

#ifdef F32
#define __USE_GNU
#define __HAVE_FLOAT32 1
#define MYFLOAT float
#define LOG(X) logf(X)
#define CEIL(X) ceilf(X)
#define EXP(X) expf(X)
#define SIN(X) sinf(X)
#define ABS(X) fabsf(X)
#define LGAMMA(X) f_lgamma(X)
#define PIGRECO 3.141592653589793238462643383279502884
#endif
  
/* type.c */
char * fstr(MYFLOAT x);
int readf(const char *str, MYFLOAT *x);

/* fft.c */
int runFFT(unsigned long long int q, unsigned long long int primroot, const char *basename, const char *datafile);

/* mmap.c */
void* mapRO_File2Array(const char *filename, int* fd, size_t arraySize, size_t typeSize);
void* mapRW_File2Array(const char* filename, int* fd, size_t arraySize, size_t typeSize);
void  unmap_File2Array(void * mmappedData, size_t filesize, int fd);
char* mallocFilename(const char* basename, const char* suffix);

#endif
